# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import sys

from wavcheck.data import FilenameTcFormat

from .timecode import DropType, FrameRate, FramesPerSec


def prompt_framerate() -> FrameRate:
    """Interactively prompts user to select a framerate."""
    print()
    print("[ 1] 23.976 non-drop    [ 7] 47.952 non-drop")
    print("[ 2] 24.000 non-drop    [ 8] 48.000 non-drop")
    print("[ 3] 25.000 non-drop    [ 9] 50.000 non-drop")
    print("[ 4] 29.970 drop        [10] 59.940 drop")
    print("[ 5] 29.970 non-drop    [11] 59.940 non-drop")
    print("[ 6] 30.000 non-drop    [12] 60.000 non-drop")
    print()

    try:
        choice = int(input(">>>>>>>>>> Select your framerate (1-12): "))
        if choice < 1 or 12 < choice:
            raise Exception()
    except Exception:
        sys.exit("[wavcheck] ERROR: Invalid framerate input")

    if choice == 1:
        return FrameRate(FramesPerSec.FPS_23_976, DropType.NON_DROP)
    elif choice == 2:
        return FrameRate(FramesPerSec.FPS_24_000, DropType.NON_DROP)
    elif choice == 3:
        return FrameRate(FramesPerSec.FPS_25_000, DropType.NON_DROP)
    elif choice == 4:
        return FrameRate(FramesPerSec.FPS_29_970, DropType.DROP)
    elif choice == 5:
        return FrameRate(FramesPerSec.FPS_29_970, DropType.NON_DROP)
    elif choice == 6:
        return FrameRate(FramesPerSec.FPS_30_000, DropType.NON_DROP)
    elif choice == 7:
        return FrameRate(FramesPerSec.FPS_47_952, DropType.NON_DROP)
    elif choice == 8:
        return FrameRate(FramesPerSec.FPS_48_000, DropType.NON_DROP)
    elif choice == 9:
        return FrameRate(FramesPerSec.FPS_50_000, DropType.NON_DROP)
    elif choice == 10:
        return FrameRate(FramesPerSec.FPS_59_940, DropType.DROP)
    elif choice == 11:
        return FrameRate(FramesPerSec.FPS_59_940, DropType.NON_DROP)
    else:
        return FrameRate(FramesPerSec.FPS_60_000, DropType.NON_DROP)


def prompt_write_framerate_file() -> bool:
    """Asks user if they want to write a FRAMERATE.txt file."""
    print(">>>>>>>>>> 📎 So I don't have to ask next time, would you like me to")
    answer = input("              write that to a FRAMERATE.txt file? [y/N] ")
    return answer.strip().lower().startswith("y")


def prompt_should_fix_umids() -> bool:
    """Asks user if they want to make all UMIDs unique."""
    print(">>>>>>>>>> 📎 Looks like you have non-unique UMIDs, would you like me to")
    answer = input("              fix them to all be unique? [y/N] ")
    return answer.strip().lower().startswith("y")


def prompt_should_append_filename_tcs() -> bool:
    """Asks user if they want to append any missing timecodes to filenames."""
    print(">>>>>>>>>> 📎 Looks like some filenames don't contain timecodes,")
    answer = input(
        "              would you like me to rename those files for you? [y/N] ")
    return answer.strip().lower().startswith("y")


def prompt_filename_suffix_format() -> FilenameTcFormat:
    """Interactively prompts user to select a filename timecode format."""
    print()
    print("How should a file BASENAME.wav starting at 01:02:03:04 be renamed?\n")
    print("[1] BASENAME TC01020304.wav")
    print("[2] BASENAME_TC01020304.wav")
    print("[3] BASENAME TC01.02.03.04.wav")
    print("[4] BASENAME_TC01.02.03.04.wav")
    print()

    try:
        choice = int(
            input(">>>>>>>>>> Select a filename timecode suffix format (1-4): "))
        if choice < 1 or 4 < choice:
            raise Exception()
    except Exception:
        sys.exit(
            "[wavcheck] ERROR: Invalid filename timecode suffix format choice")

    if choice == 1:
        return FilenameTcFormat.SPACE_NO_DOTS
    elif choice == 2:
        return FilenameTcFormat.UNDERSCORE_NO_DOTS
    elif choice == 3:
        return FilenameTcFormat.SPACE_WITH_DOTS
    else:
        return FilenameTcFormat.UNDERSCORE_WITH_DOTS



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import unittest

from .timecode import FrameRate, Timecode, parse_framerate_within, parse_timecode_str, tc_to_wall_secs, wall_secs_to_durstr, wall_secs_to_fractional_frame_idx, wall_secs_to_tc_left


def _fr(s: str) -> FrameRate:
    frame_rate = parse_framerate_within(s)
    assert frame_rate is not None
    return frame_rate


def _tc(s: str) -> Timecode:
    return parse_timecode_str(s)


class TestWallSecsToDurStr(unittest.TestCase):
    def test_works_for_positive_durations(self):
        self.assertEqual(wall_secs_to_durstr(0), "00s")
        self.assertEqual(wall_secs_to_durstr(0.49999999), "00s")
        self.assertEqual(wall_secs_to_durstr(0.5), "01s")

        self.assertEqual(wall_secs_to_durstr(59.49999999), "59s")
        self.assertEqual(wall_secs_to_durstr(59.5), "1m 00s")

        self.assertEqual(wall_secs_to_durstr(3540), "59m 00s")
        self.assertEqual(wall_secs_to_durstr(3599.49999999), "59m 59s")
        self.assertEqual(wall_secs_to_durstr(3600), "1h 00m 00s")
        self.assertEqual(wall_secs_to_durstr(3765), "1h 02m 45s")

        self.assertEqual(wall_secs_to_durstr(359999.49999999), "99h 59m 59s")
        self.assertEqual(wall_secs_to_durstr(359999.98333333), "100h 00m 00s")

    def test_works_for_negative_durations(self):
        self.assertEqual(wall_secs_to_durstr(-0.0), "00s")
        self.assertEqual(wall_secs_to_durstr(-0.49999999), "00s")
        self.assertEqual(wall_secs_to_durstr(-0.5), "(-) 01s")

        self.assertEqual(wall_secs_to_durstr(-59.49999999), "(-) 59s")

        self.assertEqual(wall_secs_to_durstr(-3540), "(-) 59m 00s")

        self.assertEqual(wall_secs_to_durstr(-3765), "(-) 1h 02m 45s")


class TimecodeTestCase(unittest.TestCase):
    def assert_tc(self, tc: Timecode, expected: str):
        self.assertEqual(str(tc), expected)


class TestTimecodeEq(TimecodeTestCase):
    def test_works(self):
        self.assertTrue(_tc("01:02:03:04") == _tc("01:02:03:04"))
        self.assertFalse(_tc("01:02:03:04") == _tc("01:02:03:05"))

        self.assertFalse(_tc("01:02:03:04") != _tc("01:02:03:04"))
        self.assertTrue(_tc("01:02:03:04") != _tc("01:02:03:05"))


class TestParseTimecodeStr(TimecodeTestCase):
    def test_works(self):
        self.assert_tc(parse_timecode_str("01020304"), "01:02:03:04")
        self.assert_tc(parse_timecode_str("  01.02.03.04   "), "01:02:03:04")
        self.assert_tc(parse_timecode_str("1_02_03_04"), "01:02:03:04")
        self.assert_tc(parse_timecode_str("1:02:03:04"), "01:02:03:04")
        self.assert_tc(parse_timecode_str("1:02:03;04"), "01:02:03:04")


class TestTcToWallSecs(TimecodeTestCase):
    def test_works(self):
        self.assertAlmostEqual(1.04, tc_to_wall_secs(
            _tc("00:00:01:02"), _fr("50.00 non-drop")))

        self.assertAlmostEqual(60.02663333, tc_to_wall_secs(
            _tc("00:00:59:29"), _fr("29.97 drop")))
        self.assertAlmostEqual(60.06, tc_to_wall_secs(
            _tc("00:01:00;02"), _fr("29.97 drop")))

        # 44:33:22:11 => 160,402 timecode seconds plus 11 frames:
        self.assertAlmostEqual(160562.86079167, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("23.976 non-drop")))
        self.assertAlmostEqual(160402.45833333, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("24.000 non-drop")))
        self.assertAlmostEqual(160402.44, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("25.000 non-drop")))
        self.assertAlmostEqual(160562.76903333, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("29.970 non-drop")))
        self.assertAlmostEqual(160402.36666667, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("30.000 non-drop")))
        self.assertAlmostEqual(160562.63139583, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("47.952 non-drop")))
        self.assertAlmostEqual(160402.22916667, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("48.000 non-drop")))
        self.assertAlmostEqual(160402.22, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("50.000 non-drop")))
        self.assertAlmostEqual(160562.58551667, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("59.940 non-drop")))
        self.assertAlmostEqual(160402.18333333, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("60.000 non-drop")))

        self.assertAlmostEqual(160402.20863333, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("29.970 drop")))
        self.assertAlmostEqual(160402.02511667, tc_to_wall_secs(
            _tc("44:33:22:11"), _fr("59.940 drop")))


class TestWallSecsToTcLeft(TimecodeTestCase):
    def test_works(self):
        self.assert_tc(wall_secs_to_tc_left(
            1.04, _fr("50.00 non-drop")), "00:00:01:02")

        self.assert_tc(wall_secs_to_tc_left(
            60.02663333, _fr("29.97 drop")), "00:00:59:28")
        self.assert_tc(wall_secs_to_tc_left(
            60.02663334, _fr("29.97 drop")), "00:00:59:29")
        self.assert_tc(wall_secs_to_tc_left(
            60.06, _fr("29.97 drop")), "00:01:00:02")

        # 44:33:22:11 => 160,402 timecode seconds plus 11 frames:

        self.assert_tc(wall_secs_to_tc_left(160562.86079167,
                       _fr("23.976 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(160562.86079167,
                       _fr("23.98 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(160402.45833334,
                       _fr("24.000 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(160402.45833334,
                       _fr("24.00 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(
            160402.44, _fr("25.000 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(
            160402.44, _fr("25.00 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(160562.76903334,
                       _fr("29.970 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(160562.76903334,
                       _fr("29.97 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(160402.36666667,
                       _fr("30.000 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(160402.36666667,
                       _fr("30.00 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(160562.63139584,
                       _fr("47.952 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(160562.63139584,
                       _fr("47.95 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(160402.22916667,
                       _fr("48.000 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(160402.22916667,
                       _fr("48.00 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(
            160402.22, _fr("50.000 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(
            160402.22, _fr("50.00 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(160562.58551667,
                       _fr("59.940 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(160562.58551667,
                       _fr("59.94 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(160402.18333334,
                       _fr("60.000 non-drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(160402.18333334,
                       _fr("60.00 non-drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(
            160402.20863334, _fr("29.970 drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(
            160402.20863334, _fr("29.97 drop")), "44:33:22:11")

        self.assert_tc(wall_secs_to_tc_left(
            160402.02511667, _fr("59.940 drop")), "44:33:22:11")
        self.assert_tc(wall_secs_to_tc_left(
            160402.02511667, _fr("59.94 drop")), "44:33:22:11")


class TestWallSecsToFractionalFrameIdx(TimecodeTestCase):
    def test_works(self):
        self.assertAlmostEqual(
            50.0,
            wall_secs_to_fractional_frame_idx(1.0, _fr("50.000 non-drop")))
        self.assertAlmostEqual(
            50.25,
            wall_secs_to_fractional_frame_idx(1.005, _fr("50.000 non-drop")))
        self.assertAlmostEqual(
            50.5,
            wall_secs_to_fractional_frame_idx(1.01, _fr("50.000 non-drop")))
        self.assertAlmostEqual(
            50.75,
            wall_secs_to_fractional_frame_idx(1.015, _fr("50.000 non-drop")))
        self.assertAlmostEqual(
            51.0,
            wall_secs_to_fractional_frame_idx(1.02, _fr("50.000 non-drop")))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import sys

_MIN_PYTHON_VER = (3, 6)
if sys.version_info < _MIN_PYTHON_VER:
    sys.exit("[wavcheck] ERROR: Requires Python %s.%s or newer" % _MIN_PYTHON_VER)


from .main import cli

if __name__ == "__main__":
    cli()



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import enum
import math
import re

_MINS_PER_HR = 60
_SECS_PER_MIN = 60


@enum.unique
class FramesPerSec(str, enum.Enum):
    FPS_23_976 = "23.976"  # 23.976023976023976...
    FPS_24_000 = "24.000"
    FPS_25_000 = "25.000"
    FPS_29_970 = "29.970"  # 29.97002997002997...
    FPS_30_000 = "30.000"
    FPS_47_952 = "47.952"  # 47.952047952047952...
    FPS_48_000 = "48.000"
    FPS_50_000 = "50.000"
    FPS_59_940 = "59.940"  # 59.94005994005994...
    FPS_60_000 = "60.000"


class _FpsConfig:
    """Exact ratio configuration for a frames per second value."""
    frames: int
    per_wall_secs: int

    def __init__(self, frames: int, per_wall_secs: int):
        self.frames = frames
        self.per_wall_secs = per_wall_secs


_FRAME_RATES: dict[FramesPerSec, _FpsConfig] = {
    FramesPerSec.FPS_23_976: _FpsConfig(24000, 1001),  # 23.976023976023976...
    FramesPerSec.FPS_24_000: _FpsConfig(24, 1),
    FramesPerSec.FPS_25_000: _FpsConfig(25, 1),
    FramesPerSec.FPS_29_970: _FpsConfig(30000, 1001),  # 29.97002997002997...
    FramesPerSec.FPS_30_000: _FpsConfig(30, 1),
    FramesPerSec.FPS_47_952: _FpsConfig(48000, 1001),  # 47.952047952047952...
    FramesPerSec.FPS_48_000: _FpsConfig(48, 1),
    FramesPerSec.FPS_50_000: _FpsConfig(50, 1),
    FramesPerSec.FPS_59_940: _FpsConfig(60000, 1001),  # 59.94005994005994...
    FramesPerSec.FPS_60_000: _FpsConfig(60, 1),
}

_DROP_FRAMES_PER_10MINS: dict[FramesPerSec, int] = {
    FramesPerSec.FPS_29_970: 18,  # First 2 frames of minutes x1, x2, ..., x9.
    FramesPerSec.FPS_59_940: 36,  # First 4 frames of minutes x1, x2, ..., x9.
}


@enum.unique
class DropType(enum.Enum):
    NON_DROP = 1
    DROP = 2

    def __str__(self) -> str:
        if self == DropType.DROP:
            return "drop"
        else:
            return "non-drop"


class FrameRate:
    """A fully-specified framerate, with FPS and drop type."""
    _fps: FramesPerSec
    _drop_type: DropType

    def __init__(self, fps: FramesPerSec, drop_type: DropType):
        if drop_type == DropType.DROP and not fps in _DROP_FRAMES_PER_10MINS:
            raise Exception(
                f"[wavcheck] FramesPerSec {fps} is not a valid drop frame standard")

        self._fps = fps
        self._drop_type = drop_type

    def fps(self) -> FramesPerSec:
        return self._fps

    def int_fps(self) -> int:
        rate_info = _FRAME_RATES[self._fps]
        return math.ceil(float(rate_info.frames) / rate_info.per_wall_secs)

    def drop_type(self) -> DropType:
        return self._drop_type

    def drop_frames_per_10mins(self) -> int:
        if self._drop_type == DropType.NON_DROP:
            return 0
        return _DROP_FRAMES_PER_10MINS[self._fps]

    def __str__(self) -> str:
        return f"{self._fps} {self._drop_type}"


_FPS_PATTERN = r"\d\d\.?\d?\d?\d?"
_NON_DROP_PATTERN = r"(?:non[-_\s]*drop|ndf?)"
_DROP_PATTERN = r"(?:drop|df?)"

# Groups: [1] FPS, [2] Drop Type.
_FRAMERATE_PATTERN = rf"({_FPS_PATTERN})\s*({_DROP_PATTERN}|{_NON_DROP_PATTERN})?"


@enum.unique
class FrameRateMatchLevel(enum.Enum):
    SEARCH_WITHIN = 1
    REQUIRE_FULL_MATCH = 2


def parse_framerate_within(s: str, level=FrameRateMatchLevel.SEARCH_WITHIN) -> FrameRate:
    """If s contains a valid framerate string, parses and returns it."""
    if level == FrameRateMatchLevel.SEARCH_WITHIN:
        match = re.search(_FRAMERATE_PATTERN, s, re.IGNORECASE)
    else:
        match = re.match(_FRAMERATE_PATTERN, s.strip(), re.IGNORECASE)
    if match is None:
        return None

    # Make sure framerate is valid.
    fps_str = match.group(1)

    fps_num = float(fps_str)
    if fps_str == "23.976" or fps_str == "23.98":
        fps = FramesPerSec.FPS_23_976
    elif fps_num == 24:
        fps = FramesPerSec.FPS_24_000
    elif fps_num == 25:
        fps = FramesPerSec.FPS_25_000
    elif fps_str == "29.970" or fps_str == "29.97":
        fps = FramesPerSec.FPS_29_970
    elif fps_num == 30:
        fps = FramesPerSec.FPS_30_000
    elif fps_str == "47.952" or fps_str == "47.95":
        fps = FramesPerSec.FPS_47_952
    elif fps_num == 48:
        fps = FramesPerSec.FPS_48_000
    elif fps_num == 50:
        fps = FramesPerSec.FPS_50_000
    elif fps_str == "59.940" or fps_str == "59.94":
        fps = FramesPerSec.FPS_59_940
    elif fps_num == 60:
        fps = FramesPerSec.FPS_60_000
    else:
        raise Exception(f"[wavcheck] Unsupported framerate fps: {fps_str}")

    # Warn user of potential ambiguity if applicable.
    fps_num_decimal_digits = max(len(fps_str) - 3, 0)  # Count after '##.'
    if (fps_num_decimal_digits == 0
            and (fps == FramesPerSec.FPS_24_000 or fps == FramesPerSec.FPS_30_000
                 or fps == FramesPerSec.FPS_48_000 or fps == FramesPerSec.FPS_60_000)):
        print()
        print((f"!! WARNING: frames per second {fps_str} is potentially ambiguous. "
               f"Specify as {fps} to avoid confusion."))
        print()

    drop_type_str = match.group(2) or ""
    drop_type = (DropType.DROP
                 if re.match(_DROP_PATTERN, drop_type_str, re.IGNORECASE)
                 else DropType.NON_DROP)

    # Reject 29.970 and 59.940 framerates that don't have an explicit drop or
    # non-drop qualifier.
    if (len(drop_type_str) == 0
            and (fps == FramesPerSec.FPS_29_970 or fps == FramesPerSec.FPS_59_940)):
        raise Exception(
            f"[wavcheck] frames per second {fps} ambiguous; specify drop or non-drop")

    return FrameRate(fps, drop_type)  # Checks combination for validity.


def wall_secs_to_durstr(wall_secs: float) -> str:
    """Converts time in wall seconds to human-readable duration string.

    Rounds fractional seconds to the nearest value (with 0.5 rounding up).
    Example output for 4994.5 seconds is "1h 23m 15s".
    """
    if not math.isfinite(wall_secs):
        raise Exception("wall_secs must be finite: " + wall_secs)

    is_negative = False
    if wall_secs < 0.0:
        wall_secs *= -1
        is_negative = True

    secs = math.floor(wall_secs + 0.5)  # Round 0.5 up.

    output = ""
    if is_negative and secs != 0:
        output += "(-) "

    hh = math.floor(secs / (_MINS_PER_HR * _SECS_PER_MIN))
    secs -= (_MINS_PER_HR * _SECS_PER_MIN) * hh

    mm = math.floor(secs / _SECS_PER_MIN)
    secs -= _SECS_PER_MIN * mm

    ss = secs

    # Output hh only if non-zero. No zero padding.
    if hh > 0:
        output += f"{hh}h "

    # Output mm only if non-zero. Zero pad if needed if there are hours.
    if hh > 0 or mm > 0:
        output += f"{mm:02}" if hh > 0 else str(mm)
        output += "m "

    # Always output ss. Zero pad if needed for 2 digits.
    output += f"{ss:02}"
    output += "s"

    return output


class Timecode:
    """Numerical timecode value with HH:MM:SS:FF."""
    hh: int
    mm: int
    ss: int
    ff: int

    def __init__(self, hh: int, mm: int, ss: int, ff: int):
        # Note: Validation happens elsewhere.
        self.hh = hh
        self.mm = mm
        self.ss = ss
        self.ff = ff

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Timecode):
            return False
        other_tc: Timecode = other
        return (self.hh == other_tc.hh
                and self.mm == other_tc.mm
                and self.ss == other_tc.ss
                and self.ff == other_tc.ff)

    def __str__(self) -> str:
        return f"{self.hh:02}:{self.mm:02}:{self.ss:02}:{self.ff:02}"


_NON_DIGIT_PATTERN = re.compile(r"[^\d]")


def parse_timecode_str(s: str) -> Timecode:
    """Parses a timecode from all the numeric digits in s."""
    digits = _NON_DIGIT_PATTERN.sub("", s)  # Remove non-digits.
    digits = f"{int(digits):08}"  # Zero pad to 8 digits.
    if len(digits) != 8:
        raise Exception(f"[wavcheck] Invalid timecode pattern '{s}'")

    hh = int(digits[0:2])
    mm = int(digits[2:4])
    ss = int(digits[4:6])
    ff = int(digits[6:8])
    return Timecode(hh, mm, ss, ff)


def _frames_per_dropped_block(fr: FrameRate) -> int:
    # A block of frames is dropped from the first second (SS=00) of 9 out of
    # every 10 minutes. For 29.97 fps, for example, there are 18 frames dropped
    # per 10 minutes, in blocks of 2 frames at a time.
    return fr.drop_frames_per_10mins() // 9


def tc_to_wall_secs(tc: Timecode, fr: FrameRate) -> float:
    """Returns timecode as (fractional) wall time in seconds from origin."""
    frame_idx = _tc_to_frame_idx(tc, fr)
    return _frame_idx_to_wall_secs(frame_idx, fr)


def _tc_to_frame_idx(tc: Timecode, fr: FrameRate) -> int:
    # Calculate first ignoring dropped frames.
    tc_total_mins = (_MINS_PER_HR * tc.hh) + tc.mm
    tc_total_secs = (_SECS_PER_MIN * tc_total_mins) + tc.ss
    frame_idx = (fr.int_fps() * tc_total_secs) + tc.ff

    # Adjust for any frame numbers that were dropped.
    if fr.drop_frames_per_10mins() > 0:
        # Frames dropped through start of HH:
        frames_dropped_per_hr = 6 * fr.drop_frames_per_10mins()
        frame_idx -= tc.hh * frames_dropped_per_hr

        # Frames dropped from start of HH to start of this 10 minute block:
        frame_idx -= (math.floor(float(tc.mm) / 10)
                      * fr.drop_frames_per_10mins())

        # Frames dropped since start of this 10 minute block:
        frame_idx -= (tc.mm % 10) * _frames_per_dropped_block(fr)

    return frame_idx


def _frame_idx_to_wall_secs(frame_idx: int, fr: FrameRate) -> float:
    rate_info = _FRAME_RATES[fr.fps()]
    return frame_idx * rate_info.per_wall_secs / float(rate_info.frames)


def wall_secs_to_tc_left(wall_secs: float, fr: FrameRate) -> Timecode:
    """Returns timecode of closest frame before or exactly equal to given wall_secs."""
    fractional_frame_idx = wall_secs_to_fractional_frame_idx(wall_secs, fr)
    frame_idx = math.floor(fractional_frame_idx)
    return _frame_idx_to_tc(frame_idx, fr)


def wall_secs_to_fractional_frame_idx(wall_secs: float, fr: FrameRate) -> float:
    """Returns (fractional) frame index of wall_secs from origin."""
    rate_info = _FRAME_RATES[fr.fps()]
    return wall_secs * rate_info.frames / rate_info.per_wall_secs


def _frame_idx_to_tc(frame_idx: int, fr: FrameRate) -> Timecode:
    if frame_idx < 0:
        raise Exception("Negative frame indexes are not supported")

    frames_per_min = fr.int_fps() * _SECS_PER_MIN
    frames_per_hr = frames_per_min * _MINS_PER_HR

    # If this is a drop frame standard, adjust for any dropped frames.
    frames_remaining = frame_idx + \
        _frames_dropped_before_frame_idx(frame_idx, fr)

    hh = math.floor(float(frames_remaining) / frames_per_hr)
    frames_remaining -= hh * frames_per_hr

    mm = math.floor(float(frames_remaining) / frames_per_min)
    frames_remaining -= mm * frames_per_min

    ss = math.floor(float(frames_remaining) / fr.int_fps())
    frames_remaining -= ss * fr.int_fps()

    ff = frames_remaining
    return Timecode(hh, mm, ss, ff)


def _frames_dropped_before_frame_idx(frame_idx: int, fr: FrameRate) -> int:
    if fr.drop_type() == DropType.NON_DROP:
        return 0

    frames_per_non_drop_min = fr.int_fps() * _SECS_PER_MIN
    frames_per_dropped_block = _frames_per_dropped_block(fr)
    frames_per_drop_min = frames_per_non_drop_min - frames_per_dropped_block

    # Count # of full blocks of 10 minutes (of timecode, not wall time).
    frames_per_10mins = 10 * frames_per_non_drop_min - fr.drop_frames_per_10mins()

    frames_remaining = frame_idx
    num_complete_10min_blocks = math.floor(
        float(frames_remaining) / frames_per_10mins)
    frames_remaining -= frames_per_10mins * num_complete_10min_blocks

    num_dropped_frames = num_complete_10min_blocks * fr.drop_frames_per_10mins()

    if frames_remaining >= frames_per_non_drop_min:
        # First minute of this 10 minute block has no dropped frames.
        frames_remaining -= frames_per_non_drop_min

        # Each complete drop minute plus the current minute drops one block of frames.
        num_complete_drop_mins = math.floor(
            float(frames_remaining) / frames_per_drop_min)
        num_dropped_frames += (num_complete_drop_mins +
                               1) * frames_per_dropped_block

    return num_dropped_frames



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import collections
import sys

from .data import KSDATAFORMAT_SUBTYPE_PCM, Context, CrossFileCheck, InternalState, SupportedFormatTag, TcConfidence, WavFileCheck, WavFileState
from .timecode import wall_secs_to_fractional_frame_idx, wall_secs_to_tc_left


def check_wav_files(ctx: Context, state: InternalState):
    """Checks WAV files for potential issues."""
    if len(state.wav_files) == 0:
        sys.exit(
            "[wavcheck] ERROR: No .wav files found (or is this a John Cage thing?)")

    # Per-file checks:
    bit_depths: set[int] = set()
    sample_rates: set[int] = set()
    umid_counts = collections.Counter()

    for filename in state.wav_files:
        wav_file = state.wav_files[filename]
        _check_wav_file(ctx, wav_file)

        bit_depths.add(wav_file.metadata.fmt_data.bit_depth)
        sample_rates.add(wav_file.metadata.fmt_data.sample_rate_hz)
        if (wav_file.metadata.bwf_data is not None
                and not WavFileCheck.MISSING_UMID in wav_file.failed_checks):
            umid_counts[wav_file.metadata.bwf_data.umid_hex] += 1

    # Cross-file checks:
    if len(bit_depths) >= 2:
        state.failed_cross_checks.append(CrossFileCheck.MULTIPLE_BIT_DEPTHS)
    if len(sample_rates) >= 2:
        state.failed_cross_checks.append(CrossFileCheck.MULTIPLE_SAMPLE_RATES)
    for umid in umid_counts:
        if umid_counts[umid] >= 2:
            state.failed_cross_checks.append(CrossFileCheck.NON_UNIQUE_UMIDS)
            break


def _check_wav_file(ctx: Context, wav_state: WavFileState):
    """Performs per-file checks for one WAV file."""

    # Basic WAV metadata checks:
    fmt_data = wav_state.metadata.fmt_data
    if fmt_data.format_tag == SupportedFormatTag.WAVE_FORMAT_EXTENSIBLE:
        if fmt_data.ext_sub_format != KSDATAFORMAT_SUBTYPE_PCM:
            wav_state.failed_checks.append(WavFileCheck.NONSTANDARD_FORMAT)
    elif fmt_data.format_tag != SupportedFormatTag.WAVE_FORMAT_PCM:
        wav_state.failed_checks.append(WavFileCheck.NONSTANDARD_FORMAT)

    if fmt_data.bit_depth < 16:
        wav_state.failed_checks.append(WavFileCheck.LOW_BIT_DEPTH)

    if fmt_data.sample_rate_hz < 44100:
        wav_state.failed_checks.append(WavFileCheck.LOW_SAMPLE_RATE)

    if wav_state.metadata.duration_secs() < 1.0:
        wav_state.failed_checks.append(WavFileCheck.VERY_SHORT_DURATION)

    # Broadcast Wave Format checks:
    bwf_data = wav_state.metadata.bwf_data
    if bwf_data is None:
        wav_state.failed_checks.append(WavFileCheck.MISSING_BWF)
        return

    if bwf_data.samples_since_origin == 0:
        wav_state.failed_checks.append(WavFileCheck.STARTS_AT_TIME_ZERO)

    start_secs = wav_state.metadata.bwf_start_time_secs(ctx.frame_rate)
    bwf_tc = wall_secs_to_tc_left(start_secs, ctx.frame_rate)
    fractional_frames = wall_secs_to_fractional_frame_idx(
        start_secs, ctx.frame_rate) % 1.0
    if fractional_frames > 0.01:
        wav_state.failed_checks.append(WavFileCheck.FRACTIONAL_FRAME_START_TC)

    if (bwf_data.version == 0 or _is_all_zeros(bwf_data.umid)):
        wav_state.failed_checks.append(WavFileCheck.MISSING_UMID)

    # Timecode in filename checks (ensure it matches BWF start time):
    tc_in_filename = wav_state.metadata.tc_in_filename
    if tc_in_filename is not None and tc_in_filename.tc != bwf_tc:
        wav_state.failed_checks.append(WavFileCheck.FILENAME_TC_MISMATCH)

    if bwf_data.version < 2:
        return

    if (bwf_data.max_dbtp >= -0.3
            or bwf_data.max_momentary_lufs >= -3.0
            or bwf_data.max_short_term_lufs >= -6.0
            or bwf_data.integrated_lufs >= -9.0):
        wav_state.failed_checks.append(WavFileCheck.UNNATURALLY_LOUD)


def _is_all_zeros(data: bytes) -> bool:
    """Returns True if data is empty or consists of all zeros."""
    for i in data:
        if i != 0:
            return False
    return True



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import enum
import pathlib

from .timecode import FrameRate, Timecode

# The number of bits in a byte.
BITS_PER_BYTE = 8


class Context:
    """Contextual options and data used across this tool."""
    dir: pathlib.Path
    verbose: bool
    frame_rate: FrameRate

    def __init__(self, dir: pathlib.Path, verbose: bool):
        self.dir = dir
        self.verbose = verbose
        self.frame_rate = None  # Filled in later.


# Length of b"RIFF", <ChunkSize>, and b"WAVE" header.
WAV_HDR_LEN_BYTES = 12


@enum.unique
class SupportedFormatTag(enum.IntEnum):
    # The standard uncompressed PCM WAV format.
    WAVE_FORMAT_PCM = 0x0001

    # For MPEG-1 (audio only); technically supported by the BWF spec, but this
    # program will still warn about.
    WAVE_FORMAT_MPEG = 0x0050

    # NOTE: NOT technically BWF spec compliant unless used with the RF64 format
    # (which also allows for very large WAV files), but still common in
    # practice.
    WAVE_FORMAT_EXTENSIBLE = 0xFFFE


# Subtype for WAVE_FORMAT_EXTENSIBLE that corresponds to WAVE_FORMAT_PCM.
KSDATAFORMAT_SUBTYPE_PCM = bytes.fromhex("0100000000001000800000AA00389B71")


class FmtMetadata:
    """Holds basic WAV metadata from the fmt chunk."""
    # For all formats:
    format_tag: int
    num_chans: int
    bit_depth: int
    sample_rate_hz: int

    # For WAVE_FORMAT_EXTENSIBLE, the SubFormat GUID bytes.
    ext_sub_format: bytes

    def __init__(self):
        self.ext_sub_format = bytes()


class BwfMetadata:
    """Holds metadata read from BWF extension chunk in WAV file."""
    description: str
    originator: str
    originator_reference: str
    origination_date: str
    origination_time: str
    samples_since_origin: int
    version: int

    coding_history: str

    # For BWF Version 1+:
    umid: bytes  # 64 bytes.
    umid_hex: str

    # For BWF Version 2+:
    integrated_lufs: float
    loudness_range_lu: float
    max_dbtp: float
    max_momentary_lufs: float
    max_short_term_lufs: float


@enum.unique
class TcConfidence(enum.Enum):
    # Series of numbers interpreted as timecode, without explicit "TC" prefix.
    IMPLICIT_NUMBERS_ONLY = 0

    # Explcit "TC" prefix before series of numbers.
    EXPLICIT_TC_PREFIX = 1


class FilenameTimecode:
    """Timecode interpreted from a filename."""
    tc: Timecode
    confidence: TcConfidence

    def __init__(self, tc: Timecode, confidence: TcConfidence):
        self.tc = tc
        self.confidence = confidence


@enum.unique
class FilenameTcFormat(enum.Enum):
    # BASENAME TC01020304.wav:
    SPACE_NO_DOTS = 1

    # BASENAME_TC01020304.wav:
    UNDERSCORE_NO_DOTS = 2

    # BASENAME TC01.02.03.04.wav:
    SPACE_WITH_DOTS = 3

    # BASENAME_TC01.02.03.04.wav:
    UNDERSCORE_WITH_DOTS = 4

    def apply(self, filename: str, tc: Timecode) -> str:
        assert filename.endswith(".wav")
        new_name = filename.removesuffix(".wav")
        new_name += "_TC" if self._use_underscore() else " TC"
        new_name += str(tc).replace(":", "." if self._use_dots() else "")
        new_name += ".wav"
        return new_name

    def _use_underscore(self) -> bool:
        return (self == FilenameTcFormat.UNDERSCORE_NO_DOTS
                or self == FilenameTcFormat.UNDERSCORE_WITH_DOTS)

    def _use_dots(self) -> bool:
        return (self == FilenameTcFormat.SPACE_WITH_DOTS
                or self == FilenameTcFormat.UNDERSCORE_WITH_DOTS)


class WavMetadata:
    """Holds metadata read from a WAV file."""
    path: pathlib.Path

    # Basic fmt chunk metadata (required for a valid WAV):
    fmt_data: FmtMetadata

    # Broadcast Wave Format (BWF) metadata (None if missing):
    bwf_data: BwfMetadata

    # The number of audio bytes in the data chunk.
    data_size_bytes: int

    # Timecode parsed from the filename (e.g. some_prefix_TC01020304.wav).
    tc_in_filename: FilenameTimecode

    def __init__(self, path: pathlib.Path):
        self.path = path
        self.fmt_data = None
        self.bwf_data = None
        self.data_size_bytes = 0

    def duration_secs(self) -> float:
        """Returns duration of this WAV file in (fractional) seconds."""
        frame_size_bytes = self.fmt_data.num_chans * \
            (self.fmt_data.bit_depth / BITS_PER_BYTE)
        num_frames = self.data_size_bytes // frame_size_bytes
        return float(num_frames) / self.fmt_data.sample_rate_hz

    def bwf_start_time_secs(self, frame_rate: FrameRate) -> float:
        """Returns start time in (fractional) seconds from origin time."""
        assert self.bwf_data is not None  # Only call if present.

        start_samples = self.bwf_data.samples_since_origin
        return float(start_samples) / self.fmt_data.sample_rate_hz


@enum.unique
class WavFileCheck(enum.IntEnum):
    NONSTANDARD_FORMAT = 1
    LOW_BIT_DEPTH = 2
    LOW_SAMPLE_RATE = 3
    VERY_SHORT_DURATION = 4
    MISSING_BWF = 5
    STARTS_AT_TIME_ZERO = 6
    FRACTIONAL_FRAME_START_TC = 7
    MISSING_UMID = 8
    UNNATURALLY_LOUD = 9
    FILENAME_TC_MISMATCH = 10


@enum.unique
class CrossFileCheck(enum.IntEnum):
    MULTIPLE_BIT_DEPTHS = 1
    MULTIPLE_SAMPLE_RATES = 2
    NON_UNIQUE_UMIDS = 3


class WavFileState:
    """Maintains per-file state throughout each step in this program."""
    metadata: WavMetadata
    failed_checks: list[WavFileCheck]

    def __init__(self):
        self.failed_checks = []


class InternalState:
    """Internal state passed across each step of this checker."""
    wav_files: dict[str, WavFileState]
    failed_cross_checks: list[WavFileCheck]

    def __init__(self):
        self.wav_files = {}
        self.failed_cross_checks = []

    def warning_count(self):
        num_warnings = 0

        for filename in self.wav_files:
            num_warnings += len(self.wav_files[filename].failed_checks)
        num_warnings += len(self.failed_cross_checks)

        return num_warnings



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import enum
import collections
import os
import secrets

from .data import Context, CrossFileCheck, InternalState, TcConfidence, WavFileCheck, WavFileState
from .prompt import prompt_filename_suffix_format, prompt_should_append_filename_tcs, prompt_should_fix_umids
from .timecode import wall_secs_to_tc_left
from .write import write_new_umid


def maybe_fix_wav_files(ctx: Context, state: InternalState):
    """If applicable, prompts user to fix certain issues with WAV files."""
    _maybe_fix_umids(ctx, state)
    _maybe_add_tc_to_filenames(ctx, state)


def _maybe_fix_umids(ctx: Context, state: InternalState):
    if CrossFileCheck.NON_UNIQUE_UMIDS not in state.failed_cross_checks:
        return

    print()
    if not prompt_should_fix_umids():
        return

    seen_hex_umids: set[str] = set()
    for filename in state.wav_files:
        wav_file = state.wav_files[filename]
        if wav_file.metadata.bwf_data is not None:
            hex_umid = wav_file.metadata.bwf_data.umid_hex
            if hex_umid in seen_hex_umids:
                # Non-unique! Fix this file.
                hex_umid = _fix_umid(wav_file, seen_hex_umids)

            seen_hex_umids.add(hex_umid)

    print(("\n[wavcheck] DONE! Now all your WAV files can be their precious, "
           "unique selves. ❄️ ❄️ ❄️\n"))


def _fix_umid(wav_file: WavFileState, seen_hex_umids: set[str]) -> str:
    umid = bytearray(wav_file.metadata.bwf_data.umid)
    assert len(umid) == 64

    # Basic UMID bytes (first 32 bytes):
    # [12] Universal Label
    # [1] L
    # [3] Instance Number
    # [16] Material Number

    # Extended UMID bytes (last 32 bytes):
    # [8] Time/Date
    # [12] Spatial Coordinates
    # [4] Country
    # [4] Org
    # [4] User

    # Fix UMID by randomly generating a new Material Number.
    while True:
        umid[13:16] = b"\x00\x00\x00"  # Instance Number 0.
        umid[16:32] = secrets.token_bytes(16)

        new_umid_hex = umid.hex().upper()
        if new_umid_hex not in seen_hex_umids:
            break

    # Write this back to the WAV file.
    write_new_umid(wav_file, umid)
    return new_umid_hex


@enum.unique
class TcFilenameStatus(enum.Enum):
    NONE_NO_BWF_START_TIME = 0
    NONE_WITH_BWF_START_TIME = 1
    IMPLICIT_MISMATCH = 2
    IMPLICIT_MATCH = 3
    EXPLICIT_MISMATCH = 4
    EXPLICIT_MATCH = 5

    def is_potentially_fixable(self) -> bool:
        return (self == TcFilenameStatus.NONE_WITH_BWF_START_TIME
                or self == TcFilenameStatus.IMPLICIT_MISMATCH)


def _tc_filename_status(wav_file: WavFileState) -> TcFilenameStatus:
    has_bwf_start_time = (wav_file.metadata.bwf_data
                          and wav_file.metadata.bwf_data.samples_since_origin != 0)
    has_mismatch = WavFileCheck.FILENAME_TC_MISMATCH in wav_file.failed_checks

    tc_in_filename = wav_file.metadata.tc_in_filename
    if tc_in_filename is None:
        return (TcFilenameStatus.NONE_WITH_BWF_START_TIME if has_bwf_start_time
                else TcFilenameStatus.NONE_NO_BWF_START_TIME)

    if tc_in_filename.confidence == TcConfidence.IMPLICIT_NUMBERS_ONLY:
        return (TcFilenameStatus.IMPLICIT_MISMATCH if has_mismatch
                else TcFilenameStatus.IMPLICIT_MATCH)
    else:
        return (TcFilenameStatus.EXPLICIT_MISMATCH if has_mismatch
                else TcFilenameStatus.EXPLICIT_MATCH)


def _maybe_add_tc_to_filenames(ctx: Context, state: InternalState):
    statuses: dict[str, TcFilenameStatus] = {}
    status_counts = collections.Counter()

    # First take stock of what types of filename timecodes were present.
    num_fixable = 0
    has_fractional_frame_errors = False
    for filename in state.wav_files:
        wav_file = state.wav_files[filename]
        status = _tc_filename_status(wav_file)
        if status.is_potentially_fixable():
            num_fixable += 1
        statuses[filename] = status
        status_counts[status] += 1
        if WavFileCheck.FRACTIONAL_FRAME_START_TC in wav_file.failed_checks:
            has_fractional_frame_errors = True

    # If there's nothing to automatically fix, bail out.
    if num_fixable == 0:
        return

    # If there were any explicit mismatches or a mixture of implicit matches +
    # mismatches, the user should fix the existing filenames first.
    if (status_counts[TcFilenameStatus.EXPLICIT_MISMATCH] >= 1
        or (status_counts[TcFilenameStatus.IMPLICIT_MISMATCH] >= 1
            and status_counts[TcFilenameStatus.IMPLICIT_MATCH] >= 1)):
        print("[wavcheck] Please fix existing filenames with timecode mismatches.")
        return

    print()

    # If there were only implicit mismatches, these may not have been
    # interpreted correctly (less confident without explicit "TC" prefix).
    if status_counts[TcFilenameStatus.IMPLICIT_MISMATCH] >= 1:
        print("              (note potential existing filename timecode problems)")

    # If there were fractional frame start times, the user also may not want to
    # rename with these timecodes.
    if has_fractional_frame_errors:
        print("              (note fractional frame start time warnings)")

    # Ask user whether to rename fixable files and in what format.
    if not prompt_should_append_filename_tcs():
        return
    format = prompt_filename_suffix_format()

    # Rename those files.
    for filename in state.wav_files:
        status = statuses[filename]
        if not status.is_potentially_fixable():
            continue

        wav_file = state.wav_files[filename]
        start_secs = wav_file.metadata.bwf_start_time_secs(ctx.frame_rate)
        start_tc = wall_secs_to_tc_left(start_secs, ctx.frame_rate)

        new_name = format.apply(filename, start_tc)
        print(f"[wavcheck] Renaming to {new_name} ...")

        new_path = wav_file.metadata.path.parent / new_name
        os.rename(wav_file.metadata.path, new_path)



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import chunk
import os
import pathlib
import re
import struct
import sys

from .data import BwfMetadata, Context, FilenameTimecode, FmtMetadata, InternalState, SupportedFormatTag, TcConfidence, WavFileState, WavMetadata, WAV_HDR_LEN_BYTES
from .prompt import prompt_framerate, prompt_write_framerate_file
from .timecode import FrameRate, FrameRateMatchLevel, parse_framerate_within, parse_timecode_str
from .write import write_framerate_file


def read_or_prompt_framerate(ctx: Context, framerate_input: str) -> FrameRate:
    """Reads framerate input via argument, file, or prompts user for it."""
    # If user provided a -f, --framerate arg input, use it.
    if len(framerate_input) >= 1:
        return _read_framerate_from_arg_or_die(ctx, framerate_input)

    # If not, look for FRAMERATE.txt (or similarly named) within dir.
    frame_rate = _maybe_read_framerate_from_default_file(ctx)
    if frame_rate is not None:
        return frame_rate

    # Otherwise, ask the user to just tell us what the framerate is.
    frame_rate = prompt_framerate()

    # And see if they want to write their choice to FRAMERATE.txt.
    if prompt_write_framerate_file():
        write_framerate_file(ctx, frame_rate)

    return frame_rate


def _read_framerate_from_arg_or_die(ctx: Context, framerate_input: str) -> FrameRate:
    """Tries to read framerate from provided arg, or exits with failure."""
    # See if argument directly specified a framerate:
    frame_rate = parse_framerate_within(
        framerate_input, FrameRateMatchLevel.REQUIRE_FULL_MATCH)
    if frame_rate is not None:
        return frame_rate

    # Otherwise, see if this argument names a file.
    # If relative, interpret relative to dir.
    path = pathlib.Path(framerate_input)
    path = path if path.is_absolute() else ctx.dir.joinpath(path)
    path = path.resolve()
    if not path.exists() or not path.is_file():
        sys.exit((f"[wavcheck] ERROR: Trying to find a framerate in {path}, "
                  "but it doesn't exist as a file"))

    return _read_framerate_from_file_or_die(path)


_FRAMERATE_LABEL_PATTERN = r"frame[-_\s]*rate"


def _maybe_read_framerate_from_default_file(ctx: Context) -> FrameRate:
    """Tries to read framerate from FRAMERATE.txt or similar in ctx.dir."""
    candidate_file: pathlib.Path = None
    with os.scandir(ctx.dir) as entries:
        for entry in entries:
            if entry.is_file() and re.search(_FRAMERATE_LABEL_PATTERN, str(entry.name), re.IGNORECASE):
                candidate_file = pathlib.Path(entry.path)
                break
    if candidate_file is None:
        return None

    return _read_framerate_from_file_or_die(candidate_file)


def _read_framerate_from_file_or_die(file: pathlib.Path):
    """Reads framerate from file (name or contents), or exits with failure."""
    print(f"[wavcheck] Looking for a framerate within {file} ...")

    # See if framerate is directly contained in the filename.
    try:
        frame_rate = parse_framerate_within(file.name)
        if frame_rate is not None:
            return frame_rate
    except Exception:
        pass

    # Search file contents.
    with open(file, "r") as f:
        # Look for a line of form "Framerate: <framerate>" (or similar).
        for line in f:
            match = re.search(_FRAMERATE_LABEL_PATTERN, line, re.IGNORECASE)
            if match:
                frame_rate = parse_framerate_within(line)
                break

        # Otherwise, see if entire file contents match.
        if frame_rate is None:
            f.seek(0)
            frame_rate = parse_framerate_within(f.read(),
                                                FrameRateMatchLevel.REQUIRE_FULL_MATCH)

    if frame_rate is None:
        sys.exit(f"[wavcheck] Unable to find valid framerate in {file}")
    return frame_rate


def read_wav_files(ctx: Context) -> InternalState:
    """Reads metadata for all WAV files in the given dir."""
    print(f"[wavcheck] Reading .wav files in '{ctx.dir}' ...")

    result = InternalState()
    with os.scandir(ctx.dir) as entries:
        for entry in entries:
            if entry.is_file() and str(entry.name).endswith(".wav"):
                if ctx.verbose:
                    print(f"[wavcheck] Reading {entry.name} ...")
                p = pathlib.Path(entry.path).resolve()
                result.wav_files[p.name] = WavFileState()
                result.wav_files[p.name].metadata = _read_wav_file(p)

    _clean_up_filename_tcs(result)
    return result


def _read_wav_file(path: pathlib.Path) -> WavMetadata:
    """Reads metadata for the given WAV file."""
    metadata = WavMetadata(path)
    metadata.tc_in_filename = _find_tc_in_filename(path.name)

    with open(path, mode="rb") as file:
        # Read all relevant RIFF chunk metadata.
        file.seek(WAV_HDR_LEN_BYTES)
        while True:
            try:
                subchunk = chunk.Chunk(file, bigendian=False)
                if subchunk.getname() == b"fmt ":
                    metadata.fmt_data = _read_fmt_metadata(subchunk)
                if subchunk.getname() == b"bext":
                    metadata.bwf_data = _read_bwf_metadata(subchunk)
                if subchunk.getname() == b"data":
                    # TODO: Support RF64 long WAV files.
                    metadata.data_size_bytes = subchunk.getsize()
                subchunk.skip()  # Advance file to next subchunk.
            except EOFError:
                break

    return metadata


_TC_NUMBERS_NO_SEPARATORS = r"\d?\d\d\d\d\d\d\d"
_TC_NUMBERS_SEPARATORS = r"\d?\d[-_. ]\d\d[-_. ]\d\d[-_. ]\d\d"
_TC_NUMBERS_FILENAME_PATTERN = rf"(?:{_TC_NUMBERS_NO_SEPARATORS}|{_TC_NUMBERS_SEPARATORS})"
_TC_EXPLICIT_FILENAME_PATTERN = re.compile(
    rf"TC[-_. ]*({_TC_NUMBERS_FILENAME_PATTERN})[^\d]", re.IGNORECASE)
_TC_IMPLICIT_FILENAME_PATTERN = re.compile(
    rf"[^\d]({_TC_NUMBERS_FILENAME_PATTERN})[^\d]")


def _find_tc_in_filename(name: str) -> FilenameTimecode:
    """Returns a Timecode if found in a known pattern within name."""
    # Prefer explicit match with TC prefix.
    match = _TC_EXPLICIT_FILENAME_PATTERN.search(name)
    if match:
        return FilenameTimecode(parse_timecode_str(match.group(1)),
                                TcConfidence.EXPLICIT_TC_PREFIX)

    # Otherwise, look for the last matching sequence of 7-8 digits.
    last_match = None
    for last_match in _TC_IMPLICIT_FILENAME_PATTERN.finditer(name):
        continue
    if last_match:
        return FilenameTimecode(parse_timecode_str(last_match.group(1)),
                                TcConfidence.IMPLICIT_NUMBERS_ONLY)

    return None  # No timecode found.


_FMT_STRUCT_PACK_FMT = (
    "<"      # Little endian.

    # Common Fields:
    "H"      # [0] FormatTag (uint16)
    "H"      # [1] NumChannels (uint16)
    "I"      # [2] SamplesPerSec (uint32)
    "I"      # [3] AvgBytesPerSec (uint32)
    "H"      # [4] BlockAlign (uint16)

    # At leats valid for WAVE_FORMAT_PCM:
    "H"      # [5] BitsPerSample (uint16)
)


_FMT_EXTENSIBLE_PACK_FMT = (
    "<"      # Little endian.
    "H"      # [0] ExtraBytes (uint16)
    "H"      # [1] ValidBitsPerSample (uint16)
    "I"      # [2] ChannelMask (uint32)
    "16s"    # [3] SubFormat GUID (byte[16])
)


def _read_fmt_metadata(fmt_chunk: chunk.Chunk) -> FmtMetadata:
    """Populates fmt metadata."""
    fmt_data = fmt_chunk.read()

    common_field_size = struct.calcsize(_FMT_STRUCT_PACK_FMT)
    fmt_fields = struct.unpack(
        _FMT_STRUCT_PACK_FMT, fmt_data[:common_field_size])

    result = FmtMetadata()
    result.format_tag = fmt_fields[0]
    result.num_chans = fmt_fields[1]
    result.bit_depth = fmt_fields[5]
    result.sample_rate_hz = fmt_fields[2]

    # For WAVE_FORMAT_EXTENSIBLE, read the extra fields.
    if result.format_tag == SupportedFormatTag.WAVE_FORMAT_EXTENSIBLE:
        ext_field_size = struct.calcsize(_FMT_EXTENSIBLE_PACK_FMT)
        ext_field_end = common_field_size + ext_field_size
        ext_fields = struct.unpack(
            _FMT_EXTENSIBLE_PACK_FMT, fmt_data[common_field_size:ext_field_end])

        result.bit_depth = ext_fields[1]  # May reduce actual bits per sample.
        result.ext_sub_format = ext_fields[3]

    return result


_BWF_STRUCT_PACK_FMT = (
    "<"      # Little endian.

    # For BWF Version 0+:
    "256s"   # [0] Descsription (char[256])
    "32s"    # [1] Originator (char[32])
    "32s"    # [2] OriginatorReference (char[32])
    "10s"    # [3] OriginationDate (char[10])
    "8s"     # [4] OriginationTime (char[8])
    "Q"      # [5] TimeReference (uint64)
    "H"      # [6] Version (uint16)

    # For BWF Version 1+:
    "64s"    # [7] UMID (byte[64])

    # For BWF Version 2+:
    "h"      # [8] LoudnessValue
    "h"      # [9] LoudnessRange
    "h"      # [10] MaxTruePeakLevel
    "h"      # [11] MaxMomentaryLoudness
    "h"      # [12] MaxShortTermLoudness

    # Reserved space for future versions:
    "180s"

    # Coding history follows, but it is a variable-length CRLF-terminated string.
)


def _read_bwf_metadata(bwf_chunk: chunk.Chunk) -> BwfMetadata:
    """Populates BWF metadata."""
    bwf_data = bwf_chunk.read()

    coding_history_offset = struct.calcsize(_BWF_STRUCT_PACK_FMT)
    bwf_fields = struct.unpack(
        _BWF_STRUCT_PACK_FMT, bwf_data[:coding_history_offset])

    result = BwfMetadata()
    result.description = _safe_str(bwf_fields[0])
    result.originator = _safe_str(bwf_fields[1])
    result.originator_reference = _safe_str(bwf_fields[2])
    result.origination_date = _safe_str(bwf_fields[3])
    result.origination_time = _safe_str(bwf_fields[4])
    result.samples_since_origin = bwf_fields[5]
    result.version = bwf_fields[6]

    if result.version >= 1:
        result.umid = bwf_fields[7]
        result.umid_hex = result.umid.hex().upper()

    if result.version >= 2:
        result.integrated_lufs = float(bwf_fields[8]) / 100.0
        result.loudness_range_lu = float(bwf_fields[9]) / 100.0
        result.max_dbtp = float(bwf_fields[10]) / 100.0
        result.max_momentary_lufs = float(bwf_fields[11]) / 100.0
        result.max_short_term_lufs = float(bwf_fields[12]) / 100.0

    result.coding_history = _safe_str(
        bwf_data[coding_history_offset:]).rstrip("\r\n")
    return result


def _safe_str(data: bytes) -> str:
    last_nonzero_pos = 0
    for idx in reversed(range(len(data))):
        if data[idx] != 0:
            last_nonzero_pos = idx
            break

    # Interpret as ISO-8859-1 for a more forgiving ASCII.
    raw_str = data[:last_nonzero_pos + 1].decode("iso-8859-1").rstrip("\0")

    # But escape non-printable characters.
    return raw_str.encode("unicode_escape").decode("iso-8859-1")


def _clean_up_filename_tcs(state: InternalState):
    found_any_explicit = False

    for filename in state.wav_files:
        filename_tc = state.wav_files[filename].metadata.tc_in_filename
        if (filename_tc is not None
                and filename_tc.confidence == TcConfidence.EXPLICIT_TC_PREFIX):
            found_any_explicit = True
            break

    # If any explicit TC-prefixed timecodes were detected in filenames, ignore
    # any implicit digit sequences that weren't TC prefixed.
    if found_any_explicit:
        for filename in state.wav_files:
            filename_tc = state.wav_files[filename].metadata.tc_in_filename
            if (filename_tc is not None
                    and filename_tc.confidence == TcConfidence.IMPLICIT_NUMBERS_ONLY):
                # Clear.
                state.wav_files[filename].metadata.tc_in_filename = None



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import argparse
import pathlib
import sys

from .check import check_wav_files
from .data import Context
from .fix import maybe_fix_wav_files
from .read import read_or_prompt_framerate, read_wav_files
from .report import print_report, print_verbose_info

__name__ = "wavcheck"
__version__ = "0.9.1"  # NOTE: Also update setup.cfg when updating version.


# Command-line arguments:
parser = argparse.ArgumentParser(
    prog=__name__,
    description=("Check WAV files in a directory for potential "
                 "Broadcast Wave Format (BWF) problems."))
parser.add_argument("-f", "--framerate",
                    help='Framerate (e.g. "23.976 non-drop") or a file containing it')
parser.add_argument("--version", action="version", version=__version__)
parser.add_argument("-v", "--verbose", action=argparse.BooleanOptionalAction,
                    help="Print verbose info for each file")
parser.add_argument("dir", nargs="?", default=".",
                    help="Directory with WAV files (defaults to current dir)")


def cli():
    """Command-line interface."""
    args = parser.parse_args()

    # Resolve input directory to find WAV files in.
    d = pathlib.Path(args.dir).resolve()
    if not d.exists():
        sys.exit(f"[wavcheck] ERROR: input dir does not exist: '{d}'")
    if not d.is_dir():
        sys.exit(f"[wavcheck] ERROR: '{d}' is not a directory")

    ctx = Context(d, args.verbose or False)

    # Determine framerate.
    ctx.frame_rate = read_or_prompt_framerate(ctx, args.framerate or "")
    print(
        f"[wavcheck] Interpreting timecodes using frame rate {ctx.frame_rate}")

    # Read all WAV files in directory and check them for issues.
    state = read_wav_files(ctx)
    print(f"[wavcheck] Read {len(state.wav_files)} .wav files")
    check_wav_files(ctx, state)

    if args.verbose:
        print_verbose_info(ctx, state)
    print_report(ctx, state)

    # If applicable, prompt user to fix any correctible problems.
    maybe_fix_wav_files(ctx, state)

    num_warnings = state.warning_count()
    if num_warnings == 0:
        print("\n[wavcheck] SUCCESS: No warnings found. Happy scoring :)\n")

    sys.exit(num_warnings)



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import collections
import sys

from .data import BwfMetadata, Context, CrossFileCheck, InternalState, KSDATAFORMAT_SUBTYPE_PCM, SupportedFormatTag, WavFileCheck, WavFileState, WavMetadata
from .timecode import wall_secs_to_durstr, wall_secs_to_fractional_frame_idx, wall_secs_to_tc_left

# TODO: Use a terminal color output library like Colorama to add colorful text
# for all print() calls.


def print_verbose_info(ctx: Context, state: InternalState):
    """Prints verbose info for each read WAV file."""
    print()
    print("================================================================================")
    print("[wavcheck] Verbose Info for Read WAV Files (-v, --verbose)")
    print("================================================================================")
    print()

    for filename in state.wav_files:
        wav_file = state.wav_files[filename]

        print(filename)
        _print_detected_timecode_in_filename(ctx, wav_file.metadata)
        print((f"  {wav_file.metadata.fmt_data.bit_depth} bit, "
               f"{wav_file.metadata.fmt_data.sample_rate_hz / 1000.0} kHz, "
               f"{wav_file.metadata.fmt_data.num_chans} channels, "
               f"{wall_secs_to_durstr(wav_file.metadata.duration_secs())}"))

        _print_verbose_bwf_info(ctx, wav_file.metadata)
        print()


def _print_detected_timecode_in_filename(ctx: Context, wav_data: WavMetadata):
    """Prints verbose info for timecode detected in filename."""
    tc_str = "(none recognized)"
    if wav_data.tc_in_filename is not None:
        tc_str = str(wav_data.tc_in_filename.tc)
    print(f"  Timecode in Filename: {tc_str}")


def _print_verbose_bwf_info(ctx: Context, wav_data: WavMetadata):
    """Prints verbose info for this BWF metadata."""
    bwf_data = wav_data.bwf_data
    if bwf_data is None:
        print("  MISSING Broadcast Wave Format (BWF) Metadata (Timecode, UMID, Loudness)")
        return

    print(
        f"  Contains Broadcast Wave Format (BWF) v{bwf_data.version} Metadata: ")

    start_samples = bwf_data.samples_since_origin
    start_secs = wav_data.bwf_start_time_secs(ctx.frame_rate)
    start_tc = wall_secs_to_tc_left(start_secs, ctx.frame_rate)

    fractional_frames_msg = ""
    fractional_frames = wall_secs_to_fractional_frame_idx(
        start_secs, ctx.frame_rate) % 1.0
    if fractional_frames > 0.01:
        fractional_frames_msg = f" and {fractional_frames:0.3f} FRACTIONAL FRAME"

    print((f"    [Time] Start: {start_samples} samples "
           f"({wall_secs_to_durstr(start_secs)} wall time) after 00:00:00:00"))
    print((f"                  (interpreted as {start_tc}{fractional_frames_msg} "
           f"in {ctx.frame_rate})"))

    print((f"    [Orig] {bwf_data.originator} {bwf_data.originator_reference} "
           f"{bwf_data.origination_date} {bwf_data.origination_time}"))
    print(f"    [Desc] {bwf_data.description}")
    print(f"    [Hist] {bwf_data.coding_history}")

    if bwf_data.version >= 1:
        print(f"    [UMID] {bwf_data.umid_hex}")
    if bwf_data.version >= 2:
        print(f"    [Loud] {_loudness_summary(bwf_data)}")


def _loudness_summary(bwf_data: BwfMetadata) -> str:
    """Returns short string summary of BWF loudness stats."""
    return (f"I {bwf_data.integrated_lufs} LUFS, "
            f"R {bwf_data.loudness_range_lu} LU, "
            f"S {bwf_data.max_short_term_lufs} LUFS, "
            f"M {bwf_data.max_momentary_lufs} LUFS, "
            f"P {bwf_data.max_dbtp} dBTP")


def print_report(ctx: Context, state: InternalState):
    """Prints summary information based on checks."""
    num_warnings = state.warning_count()
    if num_warnings == 0:
        return

    print()
    print("================================================================================")
    print("[wavcheck] WARNINGS (you might want to check)")
    print("================================================================================")
    print()

    # Cross-file checks:
    for cross_check in state.failed_cross_checks:
        _print_cross_check(cross_check, state)

    print()

    # Per-file checks:
    for filename in state.wav_files:
        if len(state.wav_files[filename].failed_checks) == 0:
            continue

        print(f"!! Warnings for {filename}:")
        for file_check in state.wav_files[filename].failed_checks:
            _print_file_check(ctx, file_check, state.wav_files[filename])
        print()

    print()


def _print_cross_check(cross_check: CrossFileCheck, state: InternalState):
    """Prints warning information for the given cross-file check."""
    if cross_check == CrossFileCheck.MULTIPLE_BIT_DEPTHS:
        bit_depths: set[int] = set()
        for filename in state.wav_files:
            bit_depths.add(
                state.wav_files[filename].metadata.fmt_data.bit_depth)
        print(f"!! Multiple bit depths found: {bit_depths}")
        return

    if cross_check == CrossFileCheck.MULTIPLE_SAMPLE_RATES:
        sample_rates: set[int] = set()
        for filename in state.wav_files:
            sample_rates.add(
                state.wav_files[filename].metadata.fmt_data.sample_rate_hz)
        print(f"!! Multiple sample rates found: {sample_rates}")
        return

    if cross_check == CrossFileCheck.NON_UNIQUE_UMIDS:
        umids = collections.defaultdict(list)
        for filename in state.wav_files:
            metadata = state.wav_files[filename].metadata
            if metadata.bwf_data is not None and metadata.bwf_data.version >= 1:
                umids[metadata.bwf_data.umid_hex].append(filename)

        for umid in umids:
            if len(umids[umid]) >= 2:
                print(
                    f"!! These files have the same UMID (pass -v or --verbose for more info):")
                print(f"    {umids[umid]}")


def _print_file_check(ctx: Context, file_check: WavFileCheck, wav_state: WavFileState):
    """Prints warning information for the given file-specific check."""
    metadata = wav_state.metadata

    if file_check == WavFileCheck.NONSTANDARD_FORMAT:
        if metadata.fmt_data.format_tag == SupportedFormatTag.WAVE_FORMAT_EXTENSIBLE:
            print(("    WAVE_FORMAT_EXTENSIBLE should be used with "
                   f"KSDATAFORMAT_SUBTYPE_PCM (0x{KSDATAFORMAT_SUBTYPE_PCM.hex().upper()}); "
                   f"found: 0x{metadata.fmt_data.ext_sub_format.hex().upper()}"))
            return
        else:
            print(
                f"    Nonstandard FormatTag: 0x{format(metadata.fmt_data.format_tag, 'X')}")
            return

    if file_check == WavFileCheck.LOW_BIT_DEPTH:
        print(f"    Low bit-depth: {metadata.fmt_data.bit_depth}")
        return

    if file_check == WavFileCheck.LOW_SAMPLE_RATE:
        print(f"    Low sample rate: {metadata.fmt_data.sample_rate_hz}")
        return

    if file_check == WavFileCheck.VERY_SHORT_DURATION:
        print(f"    Very short duration: {metadata.duration_secs()} secs")
        return

    if file_check == WavFileCheck.MISSING_BWF:
        print("    Missing Broadcast Wave Format (BWF) metadata")
        return

    if file_check == WavFileCheck.STARTS_AT_TIME_ZERO:
        print("    BWF: Starts at timecode 00:00:00:00")
        return

    if file_check == WavFileCheck.FRACTIONAL_FRAME_START_TC:
        start_secs = wav_state.metadata.bwf_start_time_secs(ctx.frame_rate)
        bwf_tc = wall_secs_to_tc_left(start_secs, ctx.frame_rate)
        fractional_frames = wall_secs_to_fractional_frame_idx(
            start_secs, ctx.frame_rate) % 1.0
        fraction_str = f"{fractional_frames:.3f}".lstrip("0")
        print(f"    BWF: Starts on fractional frame {bwf_tc}{fraction_str}")
        print(f"         (as interpreted in {ctx.frame_rate})")

    if file_check == WavFileCheck.MISSING_UMID:
        print("    BWF: Missing Unique Material Identifier (UMID)")
        return

    if file_check == WavFileCheck.UNNATURALLY_LOUD:
        print("    BWF: Has at least 1 unnaturally loud stat")
        print(f"         {_loudness_summary(metadata.bwf_data)}")
        return

    if file_check == WavFileCheck.FILENAME_TC_MISMATCH:
        start_secs = wav_state.metadata.bwf_start_time_secs(ctx.frame_rate)
        bwf_tc = wall_secs_to_tc_left(start_secs, ctx.frame_rate)
        print(
            f"    BWF: Start timecode {bwf_tc} (in {ctx.frame_rate}) doesn't match")
        print(f"         file name's TC {metadata.tc_in_filename.tc}")
        return



# SPDX-FileCopyrightText: 2022 Barndollar Music, Ltd.
#
# SPDX-License-Identifier: Apache-2.0

import chunk
import sys

from .data import Context, WavFileState, WAV_HDR_LEN_BYTES
from .timecode import FrameRate


def write_framerate_file(ctx: Context, frame_rate: FrameRate):
    """Writes the chosen frame rate to FRAMERATE.txt in ctx.dir."""
    p = ctx.dir / "FRAMERATE.txt"
    with open(p, "w") as f:
        f.write(str(frame_rate))


# The number of bytes after start of
_BWF_UMID_OFFSET = 348

# Length of UMID in bytes.
_BWF_UMID_LEN = 64


def write_new_umid(wav_file: WavFileState, new_umid: bytearray):
    """Writes new_umid to the given wav_file."""
    assert len(new_umid) == _BWF_UMID_LEN

    print(f"[wavcheck] Updating UMID for {wav_file.metadata.path.name} ...")
    with open(wav_file.metadata.path, "r+b") as file:
        file.seek(WAV_HDR_LEN_BYTES)
        subchunk: chunk.Chunk = None

        # Seek to the BWF chunk.
        while True:
            try:
                subchunk = chunk.Chunk(file, bigendian=False)
                if subchunk.getname() == b"bext":
                    break  # Found it.
                subchunk.skip()  # Advance file to next subchunk.
            except EOFError:
                break

        if subchunk is None:
            sys.exit((f"[wavcheck] BWF chunk not found in {wav_file.metadata.path}, "
                      "was it changed on disk in the middle of running this program?"))

        # Advance to the UMID field.
        subchunk.seek(_BWF_UMID_OFFSET)
        umid_pos = file.tell()

        # For sanity, verify that these bytes match the old UMID we read previously.
        reread_umid = file.read(_BWF_UMID_LEN)
        if wav_file.metadata.bwf_data.umid != reread_umid:
            sys.exit(f"[wavcheck] {wav_file.metadata.path}: "
                     f"UMID re-read {reread_umid.hex().upper()} does not match "
                     f"previously read value {wav_file.metadata.bwf_data.umid.hex().upper()}")

        # Write the new UMID value.
        file.seek(umid_pos)
        num_written = file.write(new_umid)
        assert num_written == len(new_umid)
